# helper functions for processing inputs to disperseNN2

import numpy as np
import sys
from geopy import distance
import random
import utm
import tskit
from dispersenn2.read_input import parse_provenance


# project sample locations
def project_locs(locs, fp=None):
    # projection (plus some code for calculating error)
    locs = np.array(locs)
    locs = np.array(utm.from_latlon(locs[:, 0], locs[:, 1])[0:2]) / 1000
    locs = locs.T

    # calculate extremes
    min_lat = min(locs[:, 0])
    min_long = min(locs[:, 1])
    max_lat = max(locs[:, 0])
    max_long = max(locs[:, 1])
    lat_range = max_lat - min_lat
    long_range = max_long - min_long

    # rescale lat and long to each start at 0
    locs[:, 0] = (
        1 - ((locs[:, 0] - min_lat) / lat_range)
    ) * lat_range  # "1-" to orient north-south
    locs[:, 1] = locs[:, 1] - min_long

    # reposition sample locations to random area within the map
    if fp:
        ts = tskit.load(fp)
        W = parse_provenance(ts, "W")
        left_edge = np.random.uniform(low=0, high=W - long_range)
        bottom_edge = np.random.uniform(low=0, high=W - lat_range)
        locs[:, 0] += bottom_edge
        locs[:, 1] += left_edge

    return locs


# pad locations with zeros
def pad_locs(locs, n):
    padded = np.zeros((2, n))
    n = locs.shape[1]
    padded[:, 0:n] = locs
    return padded


# pre-processing rules:
#     1 biallelic change the alelles to 0 and 1 before inputting.
#     2. no missing data: filter or impute.
#     3. ideally no sex chromosomes, and only look at one sex at a time.
def vcf2genos(vcf_path, n, num_snps, phase):
    geno_mat = []
    vcf = open(vcf_path, "r")
    for line in vcf:
        if line[0:2] == "##":
            pass
        elif line[0] == "#":
            header = line.strip().split("\t")
            if n is None:  # option for getting sample size from vcf
                n = len(header) - 9
        else:
            newline = line.strip().split("\t")
            genos = []
            for field in range(9, len(newline)):
                geno = newline[field].split(":")[0].split("/")
                geno = [int(geno[0]), int(geno[1])]
                if phase == 1:
                    genos.append(sum(geno))
                elif phase == 2:
                    genos.append(geno[0])
                    genos.append(geno[1])
                else:
                    print("problem")
                    exit()
            for i in range((n * phase) - len(genos)):  # pad with 0s
                genos.append(0)
            geno_mat.append(genos)

    # check if enough snps
    if len(geno_mat) < num_snps:
        print("not enough snps")
        exit()
    if len(geno_mat[0]) < (n * phase):
        print("not enough samples")
        exit()

    # sample snps
    geno_mat = np.array(geno_mat)
    return geno_mat[np.random.choice(geno_mat.shape[0],
                                     num_snps,
                                     replace=False), :]


# calculate isolation by distance
def ibd(genos, coords, phase, num_snps):
    # subset for n samples (avoiding padding-zeros)
    n = 0
    for i in range(genos.shape[1]):
        reverse_index = genos.shape[1] - i - 1
        if len(set(genos[:, reverse_index])) > 1:
            n += reverse_index
            break
    n += 1  # for 0 indexing
    if phase == 2:
        n = int(n / 2)
    genos = genos[
        :, 0:n*phase
    ]  # (maybe don't need to subset, as long as we know n?)

    # if collapsed genos, make fake haplotypes
    if phase == 1:
        geno_mat2 = []
        for i in range(genos.shape[1]):
            geno1, geno2 = [], []
            for s in range(genos.shape[0]):
                combined_geno = genos[s, i]
                if combined_geno == 0.0:
                    geno1.append(0)
                    geno2.append(0)
                elif combined_geno == 2:
                    geno1.append(1)
                    geno2.append(1)
                elif combined_geno == 1:
                    alleles = [0, 1]
                    # assign random allele to each haplotype
                    geno1.append(alleles.pop(random.choice([0, 1])))
                    geno2.append(alleles[0])
                else:
                    print("bug", combined_geno)
                    exit()
            geno_mat2.append(geno1)
            geno_mat2.append(geno2)
        geno_mat2 = np.array(geno_mat2)
        genos = geno_mat2.T

    # denominator for "a"
    locus_specific_denominators = np.zeros((num_snps))
    P = (n * (n - 1)) / 2  # number of pairwise comparisons
    for i1 in range(0, n - 1):
        X11 = genos[:, i1 * 2]
        X12 = genos[:, i1 * 2 + 1]
        X1_ave = (X11 + X12) / 2  # average allelic does within individual-i
        for i2 in range(i1 + 1, n):
            X21 = genos[:, i2 * 2]
            X22 = genos[:, i2 * 2 + 1]
            X2_ave = (X21 + X22) / 2
            #
            SSw = (
                (X11 - X1_ave) ** 2
                + (X12 - X1_ave) ** 2
                + (X21 - X2_ave) ** 2
                + (X22 - X2_ave) ** 2
            )
            locus_specific_denominators += SSw
    locus_specific_denominators = locus_specific_denominators / (2 * P)
    denominator = np.sum(locus_specific_denominators)

    # numerator for "a"
    gendists = []
    for i1 in range(0, n - 1):
        X11 = genos[:, i1 * 2]
        X12 = genos[:, i1 * 2 + 1]
        X1_ave = (X11 + X12) / 2  # average allelic does within individual-i
        for i2 in range(i1 + 1, n):
            X21 = genos[:, i2 * 2]
            X22 = genos[:, i2 * 2 + 1]
            X2_ave = (X21 + X22) / 2
            #
            SSw = (
                (X11 - X1_ave) ** 2
                + (X12 - X1_ave) ** 2
                + (X21 - X2_ave) ** 2
                + (X22 - X2_ave) ** 2
            )
            Xdotdot = (X11 + X12 + X21 + X22) / 4  # mean allelic dose for pair
            # a measure of between indiv
            SSb = (X1_ave - Xdotdot) ** 2 + (X2_ave - Xdotdot) ** 2
            locus_specific_numerators = ((2 * SSb) - SSw) / 4
            numerator = np.sum(locus_specific_numerators)
            a = numerator / denominator
            gendists.append(a)

    # geographic distance
    geodists = []
    for i in range(0, n - 1):
        for j in range(i + 1, n):
            d = distance.distance(coords[i, :], coords[j, :]).km
            d = np.log(d)
            geodists.append(d)

    # regression
    from scipy import stats

    geodists = np.array(geodists)
    gendists = np.array(gendists)
    b = stats.linregress(geodists, gendists)[0]
    r = stats.pearsonr(geodists, gendists)[0]
    r2 = r**2
    Nw = 1 / b
    print("IBD r^2, slope, Nw:", r2, b, Nw)


# main
def main():
    vcf_path = sys.argv[1]
    n = sys.argv[2]
    if n == "None":
        n = None
    else:
        n = int(n)
    num_snps = int(sys.argv[3])
    outname = sys.argv[4]
    phase = int(sys.argv[5])
    geno_mat = vcf2genos(vcf_path, n, num_snps, phase)
    np.save(outname + ".genos", geno_mat)


if __name__ == "__main__":
    main()



# checking over the command line params

import os


def check_params(args):
    # avoid overwriting saved weights or other output files
    if args.train is True:
        if os.path.exists(
            args.out + "/Train/disperseNN2_" + str(args.seed) + "_model.hdf5"
        ) and args.force is False:
            print("saved model with specified output name already \
                   exists. To force overwrite, use --force.")
            exit()
    if args.predict is True and args.empirical is None:
        if os.path.exists(args.out
                          + "/Test/predictions_"
                          + str(args.seed)
                          + ".txt") and args.force is False:
            print(
                "saved predictions with specified output name already exists. \
                To force overwrite, use --force."
            )
            exit()

    # other checks
    if (
        args.train is False
        and args.predict is False
        and args.preprocess is False
        and args.plot_history is False
        and args.empirical is None
    ):
        print(
            "either --help, --train, --predict, --preprocess,\
            --empirical,  or --plot_history"
        )
        exit()
    if args.train is True or args.predict is True or args.preprocess is True:
        if args.out is None:
            print("specify output directory --out")
            exit()
    if args.preprocess is True:
        if args.num_snps is None:
            print("specify num snps via --num_snps")
            exit()
        if args.n is None:
            print("specify sample size via --n")
            exit()
    if args.predict is True and args.empirical is None:
        if args.num_pred is not None:
            if args.num_pred % args.batch_size != 0:
                print(
                    "\n\npred sets each need to be divisible by batch_size; \
                    otherwise some batches will have missing data\n\n"
                )
                exit()
    if args.edge_width != "0" and args.empirical is not None:
        print(
            "can't specify edge width and empirical locations; at least not \
            currently"
        )
        exit()



# helper utils for reading in data

import numpy as np
import os


# reads a list of filepaths, stores in list
def read_list(path):
    collection = []
    with open(path) as infile:
        for line in infile:
            newline = line.strip()
            collection.append(newline)
    return collection


# reads a list of filepaths, stores in dict
def read_dict(path):
    collection, counter = {}, 0
    with open(path) as infile:
        for line in infile:
            newline = line.strip()
            collection[counter] = newline
            counter += 1
    return collection


# reads a list of floats, stores in list
def read_single_value(path):
    collection = []
    with open(path) as infile:
        for line in infile:
            with open(line.strip()) as individual_file:
                newline = float(individual_file.readline().strip())
                collection.append(newline)
    return collection


# reads a list of floats, stores in dict
def read_single_value_dict(path):
    collection, counter = {}, 0
    with open(path) as infile:
        for line in infile:
            with open(line.strip()) as individual_file:
                newline = float(individual_file.readline().strip())
                collection[counter] = newline
                counter += 1
    return collection


# numpy-load list of floats, store as list
def load_single_value(path):
    collection = []
    with open(path) as infile:
        for line in infile:
            newline = np.load(line.strip())
            collection.append(newline)
    return collection


# numpy-load list of floats, store as dict
def load_single_value_dict(path):
    collection, counter = {}, 0
    with open(path) as infile:
        for line in infile:
            newline = np.load(line.strip())
            collection[counter] = newline
            counter += 1
    return collection


# read table of lat+long coords, store in list
def read_locs(path):
    collection = []
    with open(path) as infile:
        for line in infile:
            newline = line.strip().split()
            newline = list(map(float, newline))
            collection.append(newline)
    return collection


# fill a dictionary with a single value
def fill_dict_single_value(val, reps):
    collection = {}
    for i in range(reps):
        collection[i] = val
    return collection


# convert list to keys in a dictionary, values all zero
def list2dict(mylist):
    collection = {}
    for i in range(len(mylist)):
        collection[mylist[i]] = 0
    return collection


# convert list to dictionary with ordinal keys
def dict_from_list(mylist):
    collection = {}
    for i in range(len(mylist)):
        collection[i] = mylist[i]
    return collection


# parse tree sequence provenance for sigma and map width
def parse_provenance(ts, param):
    prov = str(ts.provenance(0)).split()
    for i in range(len(prov)):
        if param + "=" in prov[i]:
            val = float(prov[i].split("=")[1].split('"')[0])
            break
    return val


# read input paths from a preprocessed, hierarchical folder
def dict_from_preprocessed(path):
    targets, genos, locs, counter = {}, {}, {}, 0
    for root, subdir, files in os.walk(path + "/Targets/"):
        if subdir == []:  # excluding the Targets/ folder itself
            for f in files:
                targetpath = os.path.join(root, f)
                genopath = targetpath.replace("Targets", "Genos").replace(
                    "target", "genos"
                )
                locpath = targetpath.replace("Targets", "Locs").replace(
                    "target", "locs"
                )
                targets[counter] = targetpath
                genos[counter] = genopath
                locs[counter] = locpath
                counter += 1
    return targets, genos, locs



#!/usr/bin/env python

# main code for disperseNN2

import os
import sys
import random
import argparse
from sklearn.model_selection import train_test_split
from dispersenn2.check_params import check_params
from dispersenn2.read_input import list2dict
from dispersenn2.read_input import read_list
from dispersenn2.read_input import read_locs
from dispersenn2.read_input import dict_from_preprocessed
from dispersenn2.process_input import project_locs, vcf2genos
import gpustat
import itertools
import numpy as np
import matplotlib.pyplot as plt


def load_dl_modules():  # load TF only if reading input is successfull
    print("loading bigger modules")
    global tf
    import tensorflow as tf
    global DataGenerator
    from dispersenn2.data_generation import DataGenerator  # (loads TF)

    return


parser = argparse.ArgumentParser()
parser.add_argument(
    "--preprocess",
    action="store_true",
    default=False,
    help="create preprocessed tensors from tree sequences",
)
parser.add_argument(
    "--train", action="store_true", default=False, help="run training pipeline"
)
parser.add_argument(
    "--predict",
    action="store_true",
    default=False,
    help="run prediction pipeline"
)
parser.add_argument(
    "--empirical",
    default=None,
    help="prefix for vcf and locs"
)
parser.add_argument(
    "--target_list", help="list of filepaths to targets (sigma).", default=None
)
parser.add_argument(
    "--tree_list",
    help="list of tree filepaths.",
    default=None
)
parser.add_argument(
    "--edge_width",
    help="crop a fixed width from each edge of the map; \
    enter 'sigma' to set edge_width equal to sigma",
    default="0",
    type=str,
)
parser.add_argument(
    "--num_snps",
    default=None,
    type=int,
    help="number of SNPs",
)
parser.add_argument(
    "--num_pred",
    default=None,
    type=int,
    help="number of datasets to predict on"
)
parser.add_argument(
    "--n",
    default=None,
    type=int,
    help="sample size",
)
parser.add_argument(
    "--mu",
    help="beginning mutation rate: mu is increased until num_snps is achieved",
    default=1e-15,
    type=float,
)
parser.add_argument(
    "--rho",
    help="recombination rate",
    default=1e-8,
    type=float)
parser.add_argument(
    "--num_samples",
    default=1,
    type=int,
    help="number of repeated samples (each of size n) from each tree sequence",
)
parser.add_argument(
    "--num_reps",
    default=1,
    type=int,
    help="number of replicate-draws from the genotype matrix of each sample",
)
parser.add_argument(
    "--hold_out",
    default=0,
    type=int,
    help="integer, the number of tree sequences to hold out for testing.",
)
parser.add_argument(
    "--validation_split",
    default=0.2,
    type=float,
    help="proportion of training set (after holding out test data) \
    to use for validation during training.",
)
parser.add_argument(
    "--batch_size",
    default=10,
    type=int,
    help="batch size for training")
parser.add_argument(
    "--max_epochs", default=1000, type=int, help="max epochs for training"
)
parser.add_argument(
    "--patience",
    type=int,
    default=100,
    help="n epochs to run the optimizer after last improvement in \
    validation loss.",
)
parser.add_argument(
    "--dropout",
    default=0,
    type=float,
    help="proportion of weights to zero at the dropout layer.",
)
parser.add_argument(
    "--recapitate",
    action="store_true",
    help="recapitate tree sequences",
    default=False,
)
parser.add_argument(
    "--skip_mutate",
    action="store_true",
    help="skip adding mutations",
    default=False,
)
parser.add_argument(
    "--out",
    help="file name stem for output",
    default=None,
)
parser.add_argument("--seed", default=None, type=int, help="random seed.")
parser.add_argument(
    "--gpu",
    default="-1",
    type=str,
    help="gpu index (uses a single GPU). To avoid GPUs, \
    skip this flag or say '-1'. \
    To use any available GPU say 'any' ",
)
parser.add_argument(
    "--plot_history",
    default=False,
    type=str,
    help="plot training history? default: False",
)
parser.add_argument(
    "--load_weights",
    default=None,
    type=str,
    help="Path to a _weights.hdf5 file to load weight from previous run.",
)
parser.add_argument(
    "--phase",
    default=1,
    type=int,
    help="1 for unknown phase, 2 for known phase",
)
parser.add_argument(
    "--polarize",
    default=2,
    type=int,
    help="2 for major/minor, 1 for ancestral/derived",
)
parser.add_argument(
    "--keras_verbose",
    default=1,
    type=int,
    help="verbose argument passed to keras in model training. \
    0 = silent. 1 = progress bars for minibatches. \
    2 = show epochs. \
    Yes, 1 is more verbose than 2. Blame keras.",
)
parser.add_argument(
    "--threads",
    default=1,
    type=int,
    help="num threads.",
)
parser.add_argument(
    "--training_mean_sd", help="sigma mean and sd from training", default=None
)
parser.add_argument(
    "--learning_rate",
    default=1e-4,
    type=float,
    help="learning rate.",
)
parser.add_argument("--grid_coarseness", help="TO DO", default=50, type=int)
parser.add_argument(
    "--sample_grid",
    help="coarseness of grid for grid-sampling",
    default=None,
    type=int
)
parser.add_argument(
    "--upsample",
    help="number of upsample layers",
    default=6,
    type=int)
parser.add_argument(
    "--pairs",
    help="number of pairs to include in the feature block",
    type=int,
    default=None,
)
parser.add_argument(
    "--pairs_encode",
    help="number of pairs (<= pairs_encode) to use for gradient \
    in the first part of the network",
    type=int,
    default=None,
)
parser.add_argument(
    "--force",
    action="store_true",
    default=False,
    help="force overwrite of existing data: \
    including existing model, or predictions"
)
args = parser.parse_args()
check_params(args)
if len(sys.argv) == 1:
    parser.print_help()
    sys.exit(1)


def load_network():
    # set seed, gpu
    if args.seed is not None:
        tf.random.set_seed(args.seed)
        tf.keras.utils.set_random_seed(args.seed)
    if args.gpu != "any":  # 'any' will search for any available GPU
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    else:
        stats = gpustat.GPUStatCollection.new_query()
        ids = map(lambda gpu: int(gpu.entry["index"]), stats)
        ratios = map(
            lambda gpu: float(gpu.entry["memory.used"])
            / float(gpu.entry["memory.total"]),
            stats,
        )
        bestGPU = min(zip(ids, ratios), key=lambda x: x[1])[0]
        os.environ["CUDA_VISIBLE_DEVICES"] = str(bestGPU)
    if args.threads > 1:
        args.threads = int(
            np.floor(args.threads / 2)
        )  # in practice it uses double the specified threads
    tf.config.threading.set_intra_op_parallelism_threads(
        args.threads
    )  # limits (and sets) threads used during training
    tf.config.threading.set_inter_op_parallelism_threads(
        args.threads
    )  # this one is needed too.

    # allocate gpu memory dynamically
    # gpu_devices = tf.config.experimental.list_physical_devices('GPU')
    # for gpu in gpu_devices:
    #     tf.config.experimental.set_memory_growth(gpu, True)

    # update conv+pool iterations based on number of SNPs
    num_conv_iterations = int(np.floor(np.log10(args.num_snps)) - 1)
    if num_conv_iterations < 0:
        num_conv_iterations = 0

    # organize pairs of individuals
    combinations = list(itertools.combinations(range(args.n), 2))
    combinations = random.sample(combinations, args.pairs)
    combinations_encode = random.sample(combinations, args.pairs_encode)
    combinations = list2dict(combinations)
    combinations_encode = list2dict(combinations_encode)

    # load inputs
    geno_input = tf.keras.layers.Input(shape=(args.num_snps, args.n))
    loc_input = tf.keras.layers.Input(shape=(2, args.n))

    # initialize shared layers
    CONV_LAYERS = []
    conv_kernal_size = 2
    pooling_size = 10
    for i in range(num_conv_iterations):
        filter_size = 20 + 44 * (i + 1)
        CONV_LAYERS.append(
            tf.keras.layers.Conv1D(
                filter_size,
                kernel_size=conv_kernal_size,
                activation="relu",
                name="CONV_" + str(i),
            )
        )
    DENSE_0 = tf.keras.layers.Dense(128, activation="relu", name="DENSE_0")
    DENSE_1 = tf.keras.layers.Dense(128, activation="relu", name="DENSE_1")

    # convolutions for each pair
    hs = []
    for comb in combinations:
        h = tf.gather(geno_input, comb, axis=2)
        locs = tf.gather(loc_input, comb, axis=2)
        d = locs[:, :, 0] - locs[:, :, 1]
        d = tf.norm(d, ord="euclidean", axis=1)
        d = tf.keras.layers.Reshape((1,))(d)
        if comb in combinations_encode:
            for i in range(num_conv_iterations):
                h = CONV_LAYERS[i](h)
                h = tf.keras.layers.AveragePooling1D(pool_size=pooling_size)(h)
            h = tf.keras.layers.Flatten()(h)
            h = tf.keras.layers.concatenate([h, d])
            h = DENSE_0(h)
        else:  # cut gradient tape on some pairs to save memory
            for i in range(num_conv_iterations):
                h = tf.stop_gradient(CONV_LAYERS[i](h))
                h = tf.keras.layers.AveragePooling1D(pool_size=pooling_size)(h)
            h = tf.keras.layers.Flatten()(h)
            h = tf.keras.layers.concatenate([h, d])
            h = tf.stop_gradient(DENSE_0(h))
        # (unindent)
        hs.append(h)
    # (unindent)
    feature_block = tf.stack(hs, axis=1)
    print("\nfeature block:", feature_block.shape)

    # apply 2d dense layer
    h = DENSE_1(feature_block)

    # flatten and final dense
    h = tf.keras.layers.Flatten()(h)
    output = tf.keras.layers.Dense(1, activation="linear")(h)

    # model overview and hyperparams
    opt = tf.keras.optimizers.Adam(learning_rate=args.learning_rate)
    model = tf.keras.Model(
        inputs=[geno_input, loc_input],
        outputs=[output],
    )
    model.compile(loss="mse", optimizer=opt)
    # model.summary()
    for v in model.trainable_variables:
        print(v.name)
    print(
        "total params:",
        np.sum([np.prod(v.shape) for v in model.trainable_variables]),
        "\n",
    )

    # load weights
    if args.predict is True:
        if args.load_weights is None:
            weights = args.out + "/Train/disperseNN2_" + \
                str(args.seed) + "_model.hdf5"
        else:
            weights = args.load_weights
        print("loading weights:", weights)
        model.load_weights(weights)

    # callbacks
    checkpointer = tf.keras.callbacks.ModelCheckpoint(
        filepath=args.out + "/Train/disperseNN2_"
        + str(args.seed) + "_model.hdf5",
        verbose=args.keras_verbose,
        save_best_only=True,
        saveweights_only=False,
        monitor="val_loss",
        period=1,
    )
    earlystop = tf.keras.callbacks.EarlyStopping(
        monitor="val_loss", min_delta=0, patience=args.patience
    )
    reducelr = tf.keras.callbacks.ReduceLROnPlateau(
        monitor="val_loss",
        factor=0.5,
        patience=int(args.patience / 10),
        verbose=args.keras_verbose,
        mode="auto",
        min_delta=0,
        cooldown=0,
        min_lr=0,
    )

    return model, checkpointer, earlystop, reducelr


def make_generator_params_dict(
    targets,
    trees,
    shuffle,
    genos,
    locs,
    empirical_locs,
):
    params = {
        "targets": targets,
        "trees": trees,
        "num_snps": args.num_snps,
        "n": args.n,
        "batch_size": args.batch_size,
        "mu": args.mu,
        "shuffle": shuffle,
        "rho": args.rho,
        "baseseed": args.seed,
        "recapitate": args.recapitate,
        "skip_mutate": args.skip_mutate,
        "edge_width": args.edge_width,
        "phase": args.phase,
        "polarize": args.polarize,
        "genos": genos,
        "locs": locs,
        "grid_coarseness": args.grid_coarseness,
        "sample_grid": args.sample_grid,
        "empirical_locs": empirical_locs,
    }
    return params


def preprocess():
    # read lists
    trees = read_list(args.tree_list)
    target_paths = read_list(args.target_list)
    total_sims = len(trees)

    # separate training and test data
    train, test = train_test_split(np.arange(total_sims),
                                   test_size=args.hold_out)

    # loop through training targets to get mean and sd
    if os.path.isfile(args.out + "/preprocess_params.npy"):
        n, num_snps, meanSig, sdSig = np.load(
            args.out + "/preprocess_params.npy")
    else:
        targets = []
        for i in train:
            with open(target_paths[i]) as infile:
                arr = np.log(float(infile.readline().strip()))
            targets.append(arr)
        meanSig = np.mean(targets)
        sdSig = np.std(targets)
        os.makedirs(args.out + "/Train", exist_ok=True)
        np.save(
            args.out + "/preprocess_params",
            [args.n, args.num_snps, meanSig, sdSig]
        )

    # make directories
    os.makedirs(os.path.join(args.out,
                             "Train/Targets",
                             str(args.seed)),
                exist_ok=True)
    os.makedirs(os.path.join(args.out,
                             "Train/Genos",
                             str(args.seed)),
                exist_ok=True)
    os.makedirs(os.path.join(args.out,
                             "Train/Locs",
                             str(args.seed)),
                exist_ok=True)
    os.makedirs(os.path.join(args.out,
                             "Test/Targets",
                             str(args.seed)),
                exist_ok=True)
    os.makedirs(os.path.join(args.out,
                             "Test/Genos",
                             str(args.seed)),
                exist_ok=True)
    os.makedirs(os.path.join(args.out,
                             "Test/Locs",
                             str(args.seed)),
                exist_ok=True)

    # process
    load_dl_modules()  # (to get data generator)
    for i in range(total_sims):
        if i in test:
            split = "Test"
        else:
            split = "Train"
        targetfile = os.path.join(
            args.out, split, "Targets", str(args.seed), str(i) + ".target"
        )
        genofile = os.path.join(
            args.out, split, "Genos", str(args.seed), str(i) + ".genos"
        )
        locfile = os.path.join(
            args.out, split, "Locs", str(args.seed), str(i) + ".locs"
        )
        if (
            os.path.isfile(genofile + ".npy") is False
            or os.path.isfile(locfile + ".npy") is False
        ):
            if args.empirical is not None:
                locs = read_locs(args.empirical + ".locs")
                if len(locs) != args.n:
                    print("length of locs file doesn't match n")
                    exit()
                locs = project_locs(locs, trees[i])
            else:
                locs = None
            params = make_generator_params_dict(
                targets=None,
                trees=None,
                shuffle=None,
                genos=None,
                locs=None,
                empirical_locs=locs,
            )
            training_generator = DataGenerator([None], **params)
            geno_mat, locs = training_generator.sample_ts(trees[i], args.seed)
            np.save(genofile, geno_mat)
            np.save(locfile, locs)
        if (
            os.path.isfile(genofile + ".npy") is True
            and os.path.isfile(locfile + ".npy") is True
        ):  # (only add target if inputs successful)
            if os.path.isfile(targetfile + ".npy") is False:
                with open(target_paths[i]) as infile:
                    target = np.log(float(infile.readline().strip()))
                target = (target - meanSig) / sdSig
                np.save(targetfile, target)

    return


def train():
    # read targets
    print("reading input paths", flush=True)
    targets, genos, locs = dict_from_preprocessed(args.out + "/Train/")
    total_sims = len(targets)

    # grab n and num_snps from preprocessed dir
    if args.training_mean_sd is None:
        args.n, args.num_snps, meanSig, sdSig = np.load(
            args.out + "/preprocess_params.npy"
        )
    else:
        args.n, args.num_snps, meanSid, sdSig = np.load(args.training_mean_sd)
    args.n, args.num_snps = int(args.n), int(args.num_snps)

    # save training params:
    if args.pairs is None:
        args.pairs = int((args.n * (args.n - 1)) / 2)
    if args.pairs_encode is None:
        args.pairs_encode = int(args.pairs)
    np.save(
        args.out + "/Train/training_params_" + str(args.seed),
        [args.seed,
         args.max_epochs,
         args.validation_split,
         args.learning_rate,
         args.pairs,
         args.pairs_encode]
    )

    # split into val,train sets
    sim_ids = np.arange(0, total_sims)
    train, val = train_test_split(sim_ids, test_size=args.validation_split)
    if (
        len(val) * args.num_samples % args.batch_size != 0
        or len(train) * args.num_samples % args.batch_size != 0
    ):
        print(
            "\n\ntrain and val sets each need to be divisible by batch_size; \
            otherwise some batches will have missing data\n\n"
        )
        exit()

    # organize "partitions" to hand to data generator
    partition = {}
    partition["train"] = list(train)
    partition["validation"] = list(val)

    # initialize generators
    params = make_generator_params_dict(
        targets=targets,
        trees=None,
        shuffle=True,
        genos=genos,
        locs=locs,
        empirical_locs=None,
    )
    load_dl_modules()
    training_generator = DataGenerator(partition["train"], **params)
    validation_generator = DataGenerator(partition["validation"], **params)

    # train
    model, checkpointer, earlystop, reducelr = load_network()
    print("training!")
    model.fit(
        x=training_generator,
        epochs=args.max_epochs,
        shuffle=False,  # (redundant with shuffling inside the generator)
        verbose=args.keras_verbose,
        validation_data=validation_generator,
        callbacks=[checkpointer, earlystop, reducelr],
    )  # multi-thread via tf.config.threading.set_intra_op_parallelism_threads

    return


def predict():
    # grab mean and sd from training distribution
    if args.training_mean_sd is None:
        args.n, args.num_snps, meanSig, sdSig = np.load(
            args.out + "/preprocess_params.npy"
        )
    else:
        args.n, args.num_snps, meanSid, sdSig = np.load(args.training_mean_sd)
    args.n, args.num_snps = int(args.n), int(args.num_snps)

    # grab num pairs from saved training params
    params = np.load(args.out + "/Train/training_params_"
                     + str(args.seed) + ".npy")
    args.pairs, args.num_pairs = int(params[4]), int(params[5])

    # load inputs
    targets, genos, locs = dict_from_preprocessed(args.out + "/Test/")
    total_sims = len(targets)

    # organize "partition" to hand to data generator
    partition = {}
    if args.num_pred is None:
        args.num_pred = int(total_sims)
    simids = np.random.choice(np.arange(total_sims),
                              args.num_pred,
                              replace=False)

    # get generator ready
    params = make_generator_params_dict(
        targets=targets,
        trees=None,
        shuffle=False,
        genos=genos,
        locs=locs,
        empirical_locs=None,
    )

    # predict
    print("predicting")
    outfile = args.out + "/Test/predictions_" + str(args.seed) + ".txt"
    if os.path.isfile(outfile):
        print("pred output exists; overwriting...")
        os.remove(outfile)
    load_dl_modules()
    model, checkpointer, earlystop, reducelr = load_network()
    for b in range(
        int(np.ceil(args.num_pred / args.batch_size))
    ):  # loop to alleviate memory
        simids_batch = simids[b * args.batch_size:(b + 1) * args.batch_size]
        partition["prediction"] = np.array(simids_batch)
        generator = DataGenerator(partition["prediction"], **params)
        predictions = model.predict(generator)

        # unpack predictions
        with open(
            args.out + "/Test/predictions_" + str(args.seed) + ".txt", "a"
        ) as out_f:
            for i in range(len(predictions)):
                trueval = np.load(targets[simids_batch[i]])
                trueval = (trueval * sdSig) + meanSig
                trueval = np.exp(trueval)
                prediction = predictions[i][0]
                prediction = (prediction * sdSig) + meanSig
                prediction = np.exp(prediction)
                outline = "\t".join(map(str, [trueval, prediction]))
                print(outline, file=out_f)

    return


def empirical():
    # grab mean and sd from training distribution
    if args.training_mean_sd is None:
        args.n, args.num_snps, meanSig, sdSig = np.load(
            args.out + "/preprocess_params.npy"
        )
    else:
        args.n, args.num_snps, meanSid, sdSig = np.load(args.training_mean_sd)
    args.n, args.num_snps = int(args.n), int(args.num_snps)

    # grab num pairs from saved training params
    params = np.load(args.out + "/Train/training_params_"
                     + str(args.seed) + ".npy")
    args.pairs, args.num_pairs = int(params[4]), int(params[5])

    # project locs
    locs = read_locs(args.empirical + ".locs")
    locs = np.array(locs)
    if len(locs) != args.n:
        print("length of locs file doesn't match n")
        exit()
    locs = project_locs(locs)

    # rescale locs
    locs = np.array(locs)
    minx = min(locs[:, 0])
    maxx = max(locs[:, 0])
    miny = min(locs[:, 1])
    maxy = max(locs[:, 1])
    x_range = maxx - minx
    y_range = maxy - miny
    locs[:, 0] = (locs[:, 0] - minx) / x_range  # rescale to (0,1)
    locs[:, 1] = (locs[:, 1] - miny) / y_range
    if x_range > y_range:  # these four lines for preserving aspect ratio
        locs[:, 1] *= y_range / x_range
    elif x_range < y_range:
        locs[:, 0] *= x_range / y_range

    # load model
    load_dl_modules()
    model, checkpointer, earlystop, reducelr = load_network()

    # convert vcf to geno matrix
    for i in range(args.num_reps):
        test_genos = vcf2genos(
            args.empirical + ".vcf", args.n, args.num_snps, args.phase
        )
        # (doesn't work if sample locations are repeated)
        # ibd(test_genos, locs, args.phase, args.num_snps)
        test_genos = np.reshape(
            test_genos, (1, test_genos.shape[0], test_genos.shape[1])
        )
        test_locs = np.reshape(locs, (1, locs.shape[1], locs.shape[0]))
        prediction = model.predict([test_genos, test_locs])

        # unpack_prediction
        with open(args.out + "/empirical_" + str(args.seed) + ".txt", "a") \
             as out_f:
            prediction = prediction[0][0]
            prediction = (prediction * sdSig) + meanSig
            prediction = np.exp(prediction)
            prediction = np.round(prediction, 10)
            print(args.empirical, "rep" + str(i), prediction, file=out_f)

    return


def plot_history():
    loss, val_loss = [], [
        np.nan
    ]  # loss starts at iteration 0; val_loss starts at end of first epoch
    with open(args.plot_history) as infile:
        for line in infile:
            if "val_loss:" in line:
                endofline = line.strip().split(" loss:")[-1]
                loss.append(float(endofline.split()[0]))
                val_loss.append(float(endofline.split()[3]))
    loss.append(np.nan)  # make columns even-length
    epochs = np.arange(len(loss))
    fig = plt.figure(figsize=(4, 1.5), dpi=200)
    plt.rcParams.update({"font.size": 7})
    ax1 = fig.add_axes([0, 0, 0.4, 1])
    ax1.plot(epochs, val_loss, color="blue", lw=0.5, label="val_loss")
    ax1.set_xlabel("Epoch")
    ax1.plot(epochs, loss, color="red", lw=0.5, label="loss")
    ax1.legend()
    fig.savefig(args.plot_history + "_plot.pdf", bbox_inches="tight")


def run():

    # main #
    np.random.seed(args.seed)

    # pre-process
    if args.preprocess is True:
        print("starting pre-processing pipeline")
        preprocess()

    # train
    if args.train is True:
        print("starting training pipeline")
        train()

    # plot training history
    if args.plot_history:
        plot_history()

    # predict
    if args.predict is True:
        print("starting prediction pipeline")
        if args.empirical is None:
            print("predicting on simulated data")
            predict()
        else:
            print("predicting on empirical data")
            empirical()



# data generator code for training disperseNN2

import sys
import numpy as np
import tensorflow as tf
import msprime
import tskit
import warnings
from attrs import define
from dispersenn2.read_input import parse_provenance
import gc


@define
class DataGenerator(tf.keras.utils.Sequence):
    "Generates data for Keras"

    list_IDs: list
    targets: dict
    trees: dict
    num_snps: int
    n: int
    batch_size: int
    mu: float
    shuffle: bool
    rho: float
    baseseed: int
    recapitate: bool
    skip_mutate: bool
    edge_width: str
    phase: int
    polarize: int
    genos: dict
    locs: dict
    grid_coarseness: int
    sample_grid: int
    empirical_locs: list

    def __attrs_post_init__(self):
        "Initialize a few things"
        self.on_epoch_end()
        np.random.seed(self.baseseed)
        warnings.simplefilter(
            "ignore", msprime.TimeUnitsMismatchWarning
        )  # (recapitate step)

    def __len__(self):
        "Denotes the number of batches per epoch"
        return int(np.floor(len(self.list_IDs) / self.batch_size))

    def __getitem__(self, index):
        "Generate one batch of data"

        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size:
                               (index + 1) * self.batch_size]

        # Find list of IDs
        list_IDs_temp = [self.list_IDs[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_IDs_temp)

        return X, y

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.list_IDs))
        if self.shuffle is True:
            np.random.shuffle(self.indexes)

    def cropper(self, ts, W, sample_width, edge_width, alive_inds):
        "Cropping the map, returning individuals inside sampling window"
        cropped = []
        left_edge = np.random.uniform(
            low=edge_width, high=W - edge_width - sample_width
        )
        right_edge = left_edge + sample_width
        bottom_edge = np.random.uniform(
            low=edge_width, high=W - edge_width - sample_width
        )
        top_edge = bottom_edge + sample_width

        for i in alive_inds:
            ind = ts.individual(i)
            loc = ind.location[0:2]
            if (
                loc[0] > left_edge
                and loc[0] < right_edge
                and loc[1] > bottom_edge
                and loc[1] < top_edge
            ):
                cropped.append(i)

        return cropped

    def sample_ind(self, ts, sampled_inds, W, i, j):
        bin_size = W / self.sample_grid
        output_ind = None
        for ind in sampled_inds:
            indiv = ts.individual(ind)
            loc = indiv.location[0:2]
            if (
                loc[0] > (i * bin_size)
                and loc[0] < ((i + 1) * bin_size)
                and loc[1] > (j * bin_size)
                and loc[1] < ((j + 1) * bin_size)
            ):
                output_ind = ind
                break
        if (
            output_ind is None
        ):  # if no individuals in the current square, choose a random ind
            output_ind = np.random.choice(sampled_inds, 1, replace=False)

        return output_ind

    def unpolarize(self, snp):
        "Change 0,1 encoding to major/minor allele. Also filter no-biallelic"
        alleles = {}
        for i in range(self.n * 2):
            a = snp[i]
            if a not in alleles:
                alleles[a] = 0
            alleles[a] += 1
        if len(alleles) == 2:
            new_genotypes = []
            major, minor = list(set(alleles))  # set() gives random order
            if alleles[major] < alleles[minor]:
                major, minor = minor, major
            for i in range(self.n * 2):  # go back through, convert genotypes
                a = snp[i]
                if a == major:
                    new_genotype = 0
                elif a == minor:
                    new_genotype = 1
                new_genotypes.append(new_genotype)
        else:
            new_genotypes = False

        return new_genotypes

    def empirical_sample(self, ts, sampled_inds, n, N, W):
        locs = np.array(self.empirical_locs)
        np.random.shuffle(locs)
        indiv_dict = {}  # tracking which indivs have been picked up already
        for i in sampled_inds:
            indiv_dict[i] = 0
        keep_indivs = []
        for pt in range(n):  # for each sampling location
            dists = {}
            for i in indiv_dict:
                ind = ts.individual(i)
                loc = ind.location[0:2]
                d = ((loc[0] - locs[pt, 0]) ** 2
                     + (loc[1] - locs[pt, 1]) ** 2) ** (
                         1 / 2
                     )
                dists[d] = i  # see what I did there?
            nearest = dists[min(dists)]
            ind = ts.individual(nearest)
            loc = ind.location[0:2]
            keep_indivs.append(nearest)
            del indiv_dict[nearest]

        return keep_indivs

    def sample_ts(self, filepath, seed):
        "The meat: load in and fully process a tree sequence"

        # read input
        ts = tskit.load(filepath)
        np.random.seed(seed)

        # grab map width and sigma from provenance
        W = parse_provenance(ts, "W")
        if self.edge_width == "sigma":
            edge_width = parse_provenance(ts, "sigma")
        else:
            edge_width = float(self.edge_width)

        # recapitate
        alive_inds = []
        for i in ts.individuals():
            alive_inds.append(i.id)
        if self.recapitate == "True":
            Ne = len(alive_inds)
            if ts.num_populations > 1:
                ts = ts.simplify()  # gets rid of weird, extraneous populations
            demography = msprime.Demography.from_tree_sequence(ts)
            demography[0].initial_size = Ne
            ts = msprime.sim_ancestry(
                initial_state=ts,
                recombination_rate=self.rho,
                demography=demography,
                start_time=ts.metadata["SLiM"]["generation"],
                random_seed=seed,
            )

        # crop map
        sample_width = W - (edge_width * 2)
        sampled_inds = self.cropper(ts,
                                    W,
                                    sample_width,
                                    edge_width,
                                    alive_inds)
        if len(sampled_inds) < self.n:
            print("\tnot enough samples, killed while-loop after 100 loops",
                  flush=True)
            exit()

        # sample individuals
        if self.sample_grid is not None:
            if self.n < self.sample_grid**2:
                print("your sample grid is too fine, \
                not enough samples to fill it")
                exit()
            keep_indivs = []
            for r in range(
                int(np.ceil(self.n / self.sample_grid**2))
            ):  # sampling from each square multiple times until >= n samples
                for i in range(self.sample_grid):
                    for j in range(self.sample_grid):
                        new_guy = self.sample_ind(ts, sampled_inds, W, i, j)
                        keep_indivs.append(new_guy)
                        sampled_inds.remove(new_guy)  # avoid sampling same guy
            keep_indivs = np.random.choice(
                keep_indivs, self.n, replace=False
            )  # taking n from the >=n list
        elif self.empirical_locs is not None:
            keep_indivs = self.empirical_sample(
                ts, sampled_inds, self.n, len(sampled_inds), W
            )
        else:
            keep_indivs = np.random.choice(sampled_inds, self.n, replace=False)
        # (unindent)
        keep_nodes = []
        for i in keep_indivs:
            ind = ts.individual(i)
            keep_nodes.extend(ind.nodes)

        # simplify
        ts = ts.simplify(keep_nodes)

        # mutate
        total_snps = self.num_snps
        if self.skip_mutate is False:
            mu = float(self.mu)
            ts = msprime.sim_mutations(
                ts,
                rate=mu,
                random_seed=seed,
                model=msprime.SLiMMutationModel(type=0),
                keep=True,
            )
            counter = 0
            while ts.num_sites < (
                total_snps * 2
            ):  # extra SNPs because a few are likely  non-biallelic
                counter += 1
                mu *= 10
                ts = msprime.sim_mutations(
                    ts,
                    rate=mu,
                    random_seed=seed,
                    model=msprime.SLiMMutationModel(type=0),
                    keep=True,
                )
                if counter == 10:
                    print("\n\nsorry, Dude. Didn't generate enough snps. \n\n")
                    sys.stdout.flush()
                    exit()

        # grab spatial locations
        sample_dict = {}
        locs = []
        for samp in ts.samples():
            node = ts.node(samp)
            indID = node.individual
            if indID not in sample_dict:
                sample_dict[indID] = 0
                loc = ts.individual(indID).location[0:2]
                locs.append(loc)

        # rescale locs
        locs = np.array(locs)
        minx = min(locs[:, 0])
        maxx = max(locs[:, 0])
        miny = min(locs[:, 1])
        maxy = max(locs[:, 1])
        x_range = maxx - minx
        y_range = maxy - miny
        locs[:, 0] = (locs[:, 0] - minx) / x_range  # rescale to (0,1)
        locs[:, 1] = (locs[:, 1] - miny) / y_range
        if x_range > y_range:  # these four lines for preserving aspect ratio
            locs[:, 1] *= y_range / x_range
        elif x_range < y_range:
            locs[:, 0] *= x_range / y_range
        locs = locs.T

        # grab genos
        geno_mat0 = ts.genotype_matrix()

        # change 0,1 encoding to major/minor allele
        if self.polarize == 2:
            shuffled_indices = np.arange(ts.num_sites)
            np.random.shuffle(shuffled_indices)
            geno_mat1 = []
            snp_counter = 0
            snp_index_map = {}
            for s in range(total_snps):
                new_genotypes = self.unpolarize(geno_mat0[shuffled_indices[s]])
                if new_genotypes is not False:  # if bi-allelic, add in the snp
                    geno_mat1.append(new_genotypes)
                    snp_index_map[shuffled_indices[s]] = int(snp_counter)
                    snp_counter += 1
            while (
                snp_counter < total_snps
            ):  # likely need to replace a few non-biallelic sites
                s += 1
                new_genotypes = self.unpolarize(geno_mat0[shuffled_indices[s]])
                if new_genotypes is not False:
                    geno_mat1.append(new_genotypes)
                    snp_index_map[shuffled_indices[s]] = int(snp_counter)
                    snp_counter += 1
            geno_mat0 = []
            sorted_indices = list(snp_index_map)
            sorted_indices.sort()
            for snp in range(total_snps):
                geno_mat0.append(geno_mat1[snp_index_map[sorted_indices[snp]]])
            geno_mat0 = np.array(geno_mat0)

        # sample SNPs
        else:
            mask = [True] * total_snps + [False] * (ts.num_sites - total_snps)
            np.random.shuffle(mask)
            geno_mat0 = geno_mat0[mask, :]

        # collapse genotypes, change to minor allele dosage (e.g. 0,1,2)
        if self.phase == 1:
            geno_mat1 = np.zeros((total_snps, self.n))
            for ind in range(self.n):
                geno_mat1[:, ind] += geno_mat0[:, ind * 2]
                geno_mat1[:, ind] += geno_mat0[:, ind * 2 + 1]
            geno_mat0 = np.array(geno_mat1)  # (change variable name)

        # sample SNPs
        mask = [True] * self.num_snps + [False] * (total_snps - self.num_snps)
        np.random.shuffle(mask)
        geno_mat1 = geno_mat0[mask, :]
        geno_mat2 = np.zeros((self.num_snps, self.n * self.phase))  # pad
        geno_mat2[:, 0:self.n * self.phase] = geno_mat1

        # free memory
        del ts
        del geno_mat0
        del geno_mat1
        del mask
        gc.collect()

        return geno_mat2, locs

    def __data_generation(self, list_IDs_temp):
        "Generates data containing batch_size samples"
        X1 = np.empty((self.batch_size, self.num_snps, self.n), dtype="int8")
        X2 = np.empty((self.batch_size, 2, self.n), dtype=float)
        y = np.empty((self.batch_size,), dtype=float)
        shuffled_indices = np.arange(self.n)
        np.random.shuffle(shuffled_indices)
        for i, ID in enumerate(list_IDs_temp):
            y[i] = np.load(self.targets[ID])
            genomat = np.load(self.genos[ID])
            genomat = genomat[
                :, shuffled_indices
            ]  # shuffe augments training set; especially pairs_encode<pairs
            X1[i, :] = genomat
            X2[i, :] = np.load(self.locs[ID])
        # (unindent)
        X = [X1, X2]

        return (X, y)
